//---------------------------------------------------------------------
//  Algorithmic Conjurings @ http://www.coyotegulch.com
//  Evocosm -- An Object-Oriented Framework for Evolutionary Algorithms
//
//  runfuncopt.cpp
//---------------------------------------------------------------------
//
//  Copyright 1996, 1999, 2002, 2003, 2004, 2005, 2006, 2007 Scott Robert Ladd
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the
//      Free Software Foundation, Inc.
//      59 Temple Place - Suite 330
//      Boston, MA 02111-1307, USA.
//
//-----------------------------------------------------------------------
//
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//
//-----------------------------------------------------------------------

// Standard C++
#include <cerrno>
#include <cmath>
#include <limits>
#include <iostream>
#include <iomanip>
using namespace std;

// Function optimizer
#include "funcopt.h"
using namespace funcopt;

// openmp
#ifdef _OPENMP
#include "omp.h"
#endif

// function to initialize a solution
function_solution init_solution()
{
    // just for this routine
    static brahe_prng_state_t random;
    brahe_prng_init(&random,BRAHE_PRNG_MWC1038,0);

    // make sure the args are empty
    vector<double> solution;

    // values fall in the range [-1,1]
    for (size_t n = 0; n < 2; ++n)
        solution.push_back(brahe_prng_real2(&random) * 2.0 - 1.0);

    brahe_prng_free(&random);

    return function_solution(solution);
}

double scotts_test(vector<double> p_args)
{
    static const double PI = 3.141592653589793238;

    double result = numeric_limits<double>::min();

    // Make certain we have two, and only two values
    if (p_args.size() == 2)
    {
        // run it through the formula
        double x = p_args[0];
        double y = p_args[1];

        // automatically reject out-of-range values
        if ((x <= 1.0) && (x >= -1.0) && (y <= 1.0) && (y >= -1.0))
        {
            // the actual formula
            result = 1.0 / (0.8 + pow(x + 0.5,2.0) + 2.0 * pow(y - 0.5,2.0) - 0.3 * cos(3.0 * PI * x) - 0.4 * cos(4.0 * PI * y));

            // check for cos errors (MS uses errno for cos errors)
            if (errno != 0)
            {
                // bad cosine equals lousy fitness
                result = numeric_limits<double>::min();
                errno = 0;
            }
        }
    }

    return result;
}

int main()
{
    // create the optimizer
    function_optimizer optimizer(scotts_test,init_solution,10000,0.10F,100);

    // display OpenMP parameters if it's enabled
#ifdef _OPENMP
    cout << "\nSystem has "
         << omp_get_num_procs()
         << " processors available\nOpenMP is enabled for "
         << omp_get_max_threads()
         << " threads."
         << endl;
#endif

    // run the optimizer
    optimizer.run();

    // done -- wasn't that easy? ;}
    return 0;
}
